/*             CSE 687 Final Project --- Pointer Class            */

/*******************************************************************
 * pointer.hh                                                v1.0  *
 *                                                                 *
 * Operations: - generic pointers with safe memory allocation and  *
 *               deallocation mechnism and bound checking          *
 * part a      - Pointer class declaritions                        *
 * Language:     GNU C++ (g++)                                     *
 * System:       SUN4 SparcStation, SunOS 4.1.1 (UNIX)             *
 * Application:  CSE 687 final project                             *
 * Author:       Gang Cheng, CST 1-019,(315)443-4889(o),479-8767(h)*
 *                                                                 *
 *******************************************************************/

#ifndef Pointer_H
#define Pointer_H 1

#include <stdio.h>         // NULL
#include <iostream.h>      // cerr

extern void error(char *); // error handler

// Pointer Class definition
template<class T>
class Pointer {

 private:
  
    // structure to store general information about allocated memory
    struct P_info {
      int  size;           // memory size in number of array elements(1 if scalar)
      int  counter;        // number of pointers to this same memory
    };
  
    P_info *p_info;        // pointer to a structure storing size and count
                           //  of allocated memory (null if no memory yet)
  
    Pointer *left, *right; // double-link list to link all the pointers 
                           //  pointing to the same memory
    int index;             // current index of array for bound checking 
                           //  (0=<index<=size-1, 0 at start or scalar) 
    T *ptr;                // actual pointer to allocated memory or auto object

 public:
    
    // constructors and destructors
    Pointer();             // void (array) constructor
    Pointer(Pointer& p);   // copy constructor (for the same type)
    Pointer(T* tptr);      // promoted copy constructor(for auto variable only)
    ~Pointer();            // destructor

    // memory allocator and deallocator
    Pointer& newptr(const int n = 1); // allocator
    Pointer& deleteptr(void);         // deallocator

    // pointer arithmatic operators
    T& operator *();                  // de-reference operator
    T* operator->();                  // member de-reference operator
    T& operator[](const int i);       // subscript de-reference operator

    Pointer operator+(const int n);   // addition    operator : ptr + n
    Pointer operator-(const int n);   // subtraction operator : ptr - n
    Pointer& operator++();            // prefix addition  operator : ++ptr
    Pointer operator++(int);          // postfix addition operator : ptr++
    Pointer& operator--();            // prefix subtraction  operator: --ptr
    Pointer operator--(int);          // postfix subtraction operator: ptr--

    // assignment operators
    Pointer& operator=(Pointer&);     // assignment operator (for the same type)
    Pointer& operator=(T*);           // assignment operator (for promoted type, 
                                      //  auto variable only)

    // pointer comparison operators 
    // (NOTE: Because template with friend is not implemented on CAT, I am using 
    //        non-friend functions to define them, instead of the should-be 
    //        friend functions. This means lvalue must be a Pointer object )
    int operator==(T* tptr)      // equality operator (for promoted type)
      { return (ptr == tptr);  }       
    int operator==(Pointer& p)   // equality operator (for the same type)
      { return (ptr == p.ptr); }
    int operator!=(T* tptr)      // unequality operator (for promoted type)
      { return (ptr != tptr);  }
    int operator!=(Pointer& p)   // unequality operator (for the same type)
      { return (ptr != p.ptr); }
    int operator>(T* tptr)       // greater operator (for promoted type)
      { return (ptr > tptr);   }       
    int operator>(Pointer& p)    // greater operator (for the same type)
      { return (ptr > p.ptr);  }
    int operator<(T* tptr)       // less operator (for promoted type)
      { return (ptr < tptr);   }
    int operator<(Pointer& p)    // less operator (for the same type)
      { return (ptr < p.ptr);  }
    int operator>=(T* tptr)      // greater or equal operator (for promoted type)
      { return (ptr >= tptr);  }       
    int operator>=(Pointer& p)   // greater or equal operator (for the same type)
      { return (ptr >= p.ptr); }
    int operator<=(T* tptr)      // less or equal operator (for promoted type)
      { return (ptr <= tptr);  }
    int operator<=(Pointer& p)   // less or equal operator (for the same type)
      { return (ptr <= p.ptr); }

    // inquire functions
    int npointers(void)          // return the number of pointers to the same 
                                 //  memory (0 if no memory allocated yet)
      { return ((p_info == NULL) ? 0 : p_info->size); }
    int psize()                  // return the number of array elements (0 if 
		                 //  no memory allocated yet, 1 if scalar)
      { return ((p_info == NULL) ? 0 : p_info->counter); }

}; 

/*******************************************************************
 * pointer.hh                                                v1.0  *
 *                                                                 *
 * part b      - template implementation; a declaration also       *
 * Language:     GNU C++ (g++)                                     *
 * System:       SUN4 SparcStation, SunOS 4.1.1 (UNIX)             *
 * Application:  CSE 687 final project                             *
 * Author:       Gang Cheng, CST 1-019,(315)443-4889(o),479-8767(h)*
 *                                                                 *
 *******************************************************************/


// ----<Pointer<T> :: Pointer(void)>---- //

// void (array) constructor
template<class T>
Pointer<T> :: Pointer(void)
{
  // initilization
  p_info = NULL;
  ptr = NULL;
  left = right = NULL;
  index = 0;
}

// ----<Pointer<T> :: Pointer(Pointer& p>---- //

// copy constructor (for the same type)
template<class T>
Pointer<T> :: Pointer(Pointer& p)
{  
  // copy general information
  p_info = p.p_info;  
  index = p.index;
  ptr = p.ptr;
  left = right = NULL;
  
  // test if p is pointing to an allocated memory
  if (p.p_info != NULL) {   
    // one more pointer added to the allocated memory
    (p_info->counter)++;  // increase the number of pointers
    
    // link into the link list
    left = &p; right = p.right; 
    if (p.right != NULL) (*(p.right)).left = this;
    p.right = this;
  };
}

// ----<Pointer<T> :: Pointer(T* tptr>---- //

// promoted copy constructor(for auto variable only)
template<class T>
Pointer<T> :: Pointer(T* tptr) 
{ 
  // initilization and copy
  index = 0; p_info = NULL; 
  left = right = NULL; 
  ptr = tptr;

}

// ----<Pointer<T> :: ~Pointer(void>---- //

// destructor
template<class T>
Pointer<T> :: ~Pointer(void)
{  
  if (p_info != NULL) this->deleteptr();
}

// ----<Pointer<T>& Pointer<T> :: deleteptr(void>---- //

// memory deallocator
template<class T>
Pointer<T>& Pointer<T> :: deleteptr(void)
{ 
  // test if it is pointing to an allocated memory
  if (p_info != NULL) {
    if (--(p_info->counter) == 0) {  // no more pointer to this memory
      // delete the allocated memory
      ptr -= index;  // move to pointer to the first element 
      delete [] ptr; 
      // delete the space used for general information
      delete p_info;
    }
    else {  // there are other pointer(s) to this memory
      if (left != NULL) left->right = right;
      if (right != NULL) right->left = left;
    };
    // re-initilization
    p_info = NULL; left = right = NULL; index = 0;
  };

  return *this;
}      

// ----<T& Pointer<T> :: operator *(void)>---- //

// de-reference operator
template<class T>
T& Pointer<T> :: operator *(void) 
{   
  // bound checking
  if ((p_info == NULL && index != 0) ||
      (p_info != NULL && (index >(p_info->size -1) || index < 0))) 
    cerr << "Warning : Bad reference, outside of bound \n";
  
  return *ptr; 
}

// ----<T* Pointer<T> :: operator->(void>---- //

// member de-reference operator
template<class T>
T* Pointer<T> :: operator->(void) 
{ 
  // bound checking
  if ((p_info == NULL && index != 0) ||
      (p_info != NULL && (index >(p_info->size -1) || index < 0))) {
    cerr << "Warning : Bad reference, outside of bound \n";
    return NULL;
  }
  else  
    return ptr; 
}

// ----<T& Pointer<T> :: operator[](const int i>---- //

// subscript de-reference operator
template<class T>
T& Pointer<T> :: operator[](const int i) 
{   
  // bound checking
  if ((p_info == NULL && i != 0) ||
      (p_info != NULL && (i >(p_info->size -1) || i < 0)))
    cerr << "Warning : outside of bound \n";

  return *(ptr-index+i);
}

// ----<Pointer<T> Pointer<T> :: operator+(const int n)>---- //

// addition operator : ptr + n
template<class T>
Pointer<T> Pointer<T> :: operator+(const int n)
{  
  int i = index + n;
  
  // bound checking
  if ((p_info == NULL && i != 0) ||
      (p_info != NULL && (i >(p_info->size -1) || i < 0))) {
    cerr << "Warning : outside of bound \n";
    return NULL;
  };
  
  Pointer temp(*this);  // copy to a temporary object
  temp.index = i; (temp.ptr) += n; 
  
  return temp;
}

// ----<Pointer<T> Pointer<T> :: operator-(const int n)>---- //

// subtraction operator : ptr - n
template<class T>
Pointer<T> Pointer<T> :: operator-(const int n)
{  
  int i = index - n;

  // bound checking
  if ((p_info == NULL && i != 0) ||
      (p_info != NULL && (i >(p_info->size -1) || i < 0))) {
    cerr << "Warning : outside of bound \n";
    return NULL;
  };
  
  Pointer temp(*this); // copy to a temporiary object
  temp.index = i; (temp.ptr) -= n;
  
  return temp;
}

// ----<Pointer<T>& Pointer<T> :: operator++(void)>---- //

// prefix addition  operator : ++ptr
template<class T>
Pointer<T>& Pointer<T> :: operator++(void)
{
  index++;
  ptr++;

  // bound checking
  if ((p_info == NULL && index != 0) ||
      (p_info != NULL && (index >(p_info->size -1) || index < 0))) {
    cerr << "Warning : outside of bound \n";
    return NULL;
  };

  return *this;
}

// ----<Pointer<T> Pointer<T> :: operator++(int)>---- //

// postfix addition operator : ptr++
template<class T>
Pointer<T> Pointer<T> :: operator++(int)
{  
  Pointer save(*this); //copy to a temporiary object
  
  index++;
  ptr++;
  
  // bound checking
  if ((p_info == NULL && index != 0) ||
      (p_info != NULL && (index >(p_info->size -1) || index < 0))) {  
    cerr << "Warning : outside of bound \n";
    return NULL;
  };

  return save;
}

// ----<Pointer<T>& Pointer<T> :: operator--(void)>---- //

// prefix subtraction  operator: --ptr
template<class T>
Pointer<T>& Pointer<T> :: operator--(void)
{  
  index--;
  ptr--;

  // bound checking
  if ((p_info == NULL && index != 0) ||
      (p_info != NULL && (index >(p_info->size -1) || index < 0))) {  
    cerr << "Warning : outside of bound \n";
    return NULL;
  };

  return *this;
}

// ----<Pointer<T> Pointer<T> :: operator--(int)>---- //

// postfix subtraction operator: ptr--
template<class T>
Pointer<T> Pointer<T> :: operator--(int)
{  
  Pointer save(*this);  //copy to a temporiary object
  
  index--;
  ptr--;
  
  // bound checking
  if ((p_info == NULL && index != 0) ||
      (p_info != NULL && (index >(p_info->size -1) || index < 0))) {  
    cerr << "Warning : outside of bound \n";
    return NULL;
  };

  return save;
}

// ----<Pointer<T>& Pointer<T> :: operator=(Pointer& p)>---- //

// assignment operator (for the same type)
template<class T>
Pointer<T>& Pointer<T> :: operator=(Pointer& p)
{  
  // release the memory previously pointed to
  if (p_info != NULL) this->deleteptr();
  
  // copy general information
  p_info = p.p_info;
  ptr = p.ptr;
  index = p.index;
  
  // test if p is pointing to an allocated memory
  if (p.p_info != NULL) {   
    (p_info->counter)++;
    left = &p; right = p.right; 
    if (p.right != NULL) (*(p.right)).left = this;
    p.right = this;
  };
  
  return *this;
}

// ----<Pointer<T>& Pointer<T> :: operator=(T* tptr)>---- //

// assignment operator (for promoted type, 
//  auto variable only)
template<class T>
Pointer<T>& Pointer<T> :: operator=(T* tptr)
{
  // release the memory previously pointed to
  if (p_info != NULL) this->deleteptr();

  ptr = tptr; index = 0; p_info = NULL;
  left = right = NULL;
  
  return *this;
}

// ----<Pointer<T>& Pointer<T> :: newptr(const int n = 1)>---- //

// memory allocator
template<class T>
Pointer<T>& Pointer<T> :: newptr(const int n = 1)
{
  // release the memory currently pointing to
  if (p_info != NULL) this->deleteptr();  
  
  if (n < 1) error("Error: too small size\n");

  // allocate the object(s) memory and memory for
  //  general information 
  if ((ptr=new T[n]) == NULL || (p_info=new P_info) == NULL)
    error("Error: exhaut system memory\n");

  // initilization
  p_info->size = n;
  p_info->counter = 1;
  left = right = NULL;
  index = 0;
  
  return *this;
}

#endif Pointer_H


/*             CSE 687 Final Project --- Pointer Class 
#####################################################################
#  This is the rewritten program of Project 3 "Sparse Matrix Class" #
#  with all pointers using Pointer Class                            #
#                                                                   #
#  Almost all member functions of Pointer class are used, except:   #
#   operator [], ++, --, >, <, >=, <=, and                          #
#   inquire member functions npointers(), psize()                   #
#                                                                   #
#  (some of the missing functions are tested in the main program)   #
#                                                                   #
#####################################################################

            Student Name: Gang Cheng SSN: 223-61-3010               */

/*******************************************************************
 * matrix.hh ---- header file for the module of the member & frind *
 *		  functions for class Matrix                       *
 *******************************************************************/

#ifndef Matrix_HH
#define Matrix_HH

#include <iostream.h>     // cin, cout, cerr
#include <stdlib.h>       // exit
#include <string.h>       // strlen, strcat, strcpy, strncpy 
//#include "pointer.hh"     // Pointer class

// link list class 
struct llist {
	int id;          // connection node id
	double value;    // connection weight
	Pointer<llist> next;          // link to next connection list
	Pointer<llist> search(int id_index); // search for link-cell
      	void deallocate_llist(void);  // link-cells destructor 
};

// graph node class
struct node {
	int id;                       // node id
	Pointer<llist> link;          // link to connections
	Pointer<node> next;           // link to next node
	void deallocate_node(void);   // nodes destructor	
};

// Matrix class 
class Matrix {
   private:  
      int size;      // matrix size
      Pointer<llist> lptr;   // link list -- connection for nodes
      Pointer<node> nhead, nptr;  // top pointer & working pointer

   public:
      	// Matrix m[10]  - define arrays of objects	
	Matrix(void) { size = 0; lptr = NULL; nhead = nptr = NULL; } 
	Matrix(int msize);  // constructor
	Matrix(Matrix& m);  // Matrix m2 = m1 - copy constructor
	void deallocate_matrix();   // explicit matrix destructor
	~Matrix(void) { this->deallocate_matrix(); } // destructor
	// unary operator - m3 = m2 = m1	 
	Matrix& operator= (Matrix& m); 
	// binary operator + --- addition of two matrixes
	friend Matrix operator+ (Matrix& m1, Matrix& m2);
	// binary operator * --- multiplication of two matrixes 
	friend Matrix operator* (Matrix& m1, Matrix& m2);
 	// overloading istream operator >>
	friend istream& operator>> (istream& istrm, Matrix& m); 
 	// overloading ostream operator << 
	friend ostream& operator<< (ostream& ostrm, Matrix& m);
	int matrix_size(void) { return size; } // matrix size 
};	

#endif Matrix_HH

/*             CSE 687 Final Project --- Pointer Class 
#####################################################################
#  This is the rewritten program of Project 3 "Sparse Matrix Class" #
#  with all pointers using Pointer Class                            #
#                                                                   #
#  Almost all member functions of Pointer class are used, except:   #
#   operator [], ++, --, >, <, >=, <=, and                          #
#   inquire member functions npointers(), psize()                   #
#                                                                   #
#  (some of the missing functions are tested in the main program)   #
#                                                                   #
##################################################################### 
*/

/*******************************************************************
 * Matrix.cc ---- the module of member & frind functions           *
 *                for class Matrix                                 *
 *******************************************************************/

//#include "matrix.hh"
#include <stdio.h> // NULL
extern void error(char *);      // error handler

///////////// class Matrix member functions ///////////////

// ----- < Matrix :: Matrix(int msize) > --------------//

// - constructor: allocate space for nodes
Matrix :: Matrix(int msize)
{
#define MAXSIZE 5000
  if (msize < 0 || msize >MAXSIZE) 
    error("error: too big or small matrix size\n");
  
  Pointer<node> ntmp = NULL;
  nhead = nptr = NULL;
  size = 0;
  while (size < msize) {
    size++;
    // allocate a new node
    nptr.newptr();
    
    // initialize
    nptr->link = NULL;
    nptr->next = NULL; 
    nptr->id = size;

    if (ntmp == NULL) nhead = nptr; // pointer to the first node
    else ntmp->next = nptr;
    ntmp = nptr;
  };
#undef MAXSIZE
}

// ----< Matrix :: Matrix(Matrix& m) >----------------- //

// - copy constructs Matrix object from an existing object
Matrix :: Matrix(Matrix& m)
{
  Pointer<node> ntmp = NULL;
  nhead = nptr = NULL;
  m.nptr = m.nhead;
  size = m.size;
  int n = m.size;

  while (n-- != 0) {

    // allocate a new node 
    nptr.newptr();
        
    // initialize
    nptr->link = NULL;
    nptr->next = NULL; 
    
    // copy node information
    nptr->id = (m.nptr)->id;
   
    // allocate and copy connctions for a node
    Pointer<llist> ltmp = NULL;
    m.lptr = (m.nptr)->link;

    while (m.lptr != NULL) {
      // allocate a new link-cell
      lptr.newptr();
      lptr->next = NULL; // initialize 

      // copy link-cell information
      lptr->id = (m.lptr)->id;
      lptr->value = (m.lptr)->value;
      if (ltmp != NULL) // first cell is a special case
	ltmp->next = lptr;
      else 
	nptr->link = lptr;
      ltmp = lptr;
      m.lptr = (m.lptr)->next;
    };

    m.nptr = (m.nptr)->next;
    if (ntmp == NULL) nhead = nptr; // pointer to the first node
    else ntmp->next = nptr;
    ntmp = nptr;
  };
}

// ----< Matrix& Matrix :: operator= (Matrix& m) >--------- //

// - sets Matrix object equal to object m
Matrix& Matrix :: operator= (Matrix& m)
{ 
  if (m.size != size) error("error: incorrect matrix size\n"); 

  m.nptr = m.nhead;
  nptr = nhead;
  int n = size;

  while (n-- != 0) {
    if (m.nptr == NULL || nptr == NULL)  // bound checking
      error("error: fatal error\n");

    // copy node information
    nptr->id = (m.nptr)->id;

    // deallocate space of m's previous connection, if any
    (nptr->link)->deallocate_llist();

    // allocate and copy connctions for a node
    Pointer<llist> ltmp = NULL;
    m.lptr = (m.nptr)->link;
    while (m.lptr != NULL) {
      // allocate a new link-cell
      lptr.newptr();
      lptr->next = NULL; // initialize 

      // copy link-cell information
      lptr->id = (m.lptr)->id;
      lptr->value = (m.lptr)->value;
      if (ltmp != NULL) // first cell is a special case
	ltmp->next = lptr;
      else 
	nptr->link = lptr;
      ltmp = lptr;
      m.lptr = (m.lptr)->next;
    };

    m.nptr = (m.nptr)->next;
    nptr = nptr->next;
  };
  return *this;
}

// ----<Matrix operator+ (Matrix& m1, Matrix& m2)>------------- //

// - addition of two Matrixs
Matrix operator+ (Matrix& m1, Matrix& m2)
{
  if (m1.size != m2.size) error("error: incorrect matrix size\n"); 

  Matrix temp(m1.size);  // use togather with unary operator = which
                         // copies this temporary immediately exit 
                         // from this function
  temp = m1;   // make temp identical to m1
  m2.nptr = m2.nhead;
  temp.nptr = temp.nhead;
  int n = m2.size;

  while (n-- != 0) {
    Pointer<llist> ltmp;
    m2.lptr = (m2.nptr)->link;

    while (m2.lptr != NULL) {
      // search for a connection 
      if ((ltmp = ((temp.nptr)->link)->search((m2.lptr)->id)) != NULL)
	ltmp->value += (m2.lptr)->value;   // addition operation
      else {   // insert a new connection
	ltmp.newptr();

	// copy connection information
	ltmp->id = (m2.lptr)->id;
	ltmp->value = (m2.lptr)->value;

	// link into new matrix
	ltmp->next = (temp.nptr)->link;
	(temp.nptr)->link = ltmp;
      };
      m2.lptr = (m2.lptr)->next;
    };
    temp.nptr = (temp.nptr)->next;
    m2.nptr = (m2.nptr)->next;
  };
  return temp;
}

// ----<Matrix operator* (Matrix& m1, Matrix& m2)>------------- //

// - multiplication of two Matrixs
Matrix operator* (Matrix& m1, Matrix& m2)
{
  
  if (m1.size != m2.size) error("error: incorrect matrix size\n"); 

  Matrix temp(m1.size);  // use togather with unary operator = which
                         // copies this temporary immediately exit 
                         // from this function

  m1.nptr = m1.nhead;
  temp.nptr = temp.nhead;
  (temp.nptr)->link = NULL;
  (temp.nptr)->id = (m1.nptr)->id;

  while (m1.nptr != NULL) {
    int m = 1;

    while (m <= m1.size) { 
      Pointer<llist> ltmp;
      m1.lptr = (m1.nptr)->link;
      double sum = 0.0;

      while (m1.lptr != NULL) {
	int n2 = (m1.lptr)->id;
	m2.nptr = m2.nhead;   
	while (m2.nptr != NULL && (m2.nptr)->id != n2) 
	  m2.nptr = (m2.nptr)->next;
	if (m2.nptr == NULL) error("fatal error");

	// search for a connection 
	if ((ltmp = ((m2.nptr)->link)->search(m)) != NULL)
	  sum += ((m1.lptr)->value)*(ltmp->value);   // multiplication  
	                                             // operation
	m1.lptr = (m1.lptr)->next;
      };

      if (sum != 0.0) { // insert a new connection
	ltmp.newptr();

	// copy connection information
	ltmp->id = m;
	ltmp->value = sum;
	
	// link into new matrix
	ltmp->next = (temp.nptr)->link;
	(temp.nptr)->link = ltmp;
      };
      m++;
    };
    temp.nptr = (temp.nptr)->next;
    m1.nptr = (m1.nptr)->next;
  };
  return temp;
}


// ----<istream& operator>> (istream& istrm, Matrix& m)>----- //

// - overloading istream operator >>
istream& operator>> (istream& istrm, Matrix& m)
{
  #define MAXSIZE 5000   // maxium size of a matrix
  m.deallocate_matrix();

  int n = 0;
  char ch = ' ';
  Pointer<node> ntmp = NULL;
  m.size = 0;
  
  while (istrm) {
    if (! (istrm >> n)) continue; // skip blank lines on the bottom of file
    m.size++;
    if (n != m.size) error("error: incorrect input format(1)");

    // allocate a new node
    (m.nptr).newptr();
    
    (m.nptr)->id = n;
    (m.nptr)->link = NULL;
    istrm >>ch;
    if (ch == ':') {
      istrm >> ch;

      // input and create link_cells
      while (ch != ';') {
	istrm.putback(ch);

	// allocate a new link-cell
	(m.lptr).newptr();

	istrm >>(m.lptr)->id >> ch;
	if (ch != 'w') error("error: incorrect input format");
	istrm >>(m.lptr)->value >>ch;
	
	// insert into links
	(m.lptr)->next = (m.nptr)->link;
	(m.nptr)->link = m.lptr;
      };
    }
    else error("error: incorrect input format");

    // initialize node
    m.nptr->next = NULL; 

    // pointer to the first node
    if (ntmp == NULL) m.nhead = m.nptr; 
    else ntmp->next = m.nptr;
    ntmp = m.nptr;
    
  };

  if (!istrm.eof() || istrm.bad()) 
    error("error: incorrect input format");    

  return istrm;

  #undef MAXSIZE 
}

// ----<istream& operator<< (istream& istrm, Matrix& m)>---- //

// - overloading istream operator <<
ostream& operator<< (ostream& ostrm, Matrix& m)
{
  #define ZERO '0'          // print char. for symbol zero
  Pointer<node> ntmp = NULL;
  int n = m.size;

  ostrm <<"   ";
  ostrm.width(10);

  for (n = 1; n <= m.size; ostrm << n++) ostrm.width(10);
  ostrm << "\n   ";
  for (n = m.size ; (n--)!=0; ostrm << "..") ostrm.width(10);
  ostrm << '\n';
  
  m.nptr = m.nhead;
  while (m.nptr != NULL) {
    ostrm.width(4);
    ostrm << (m.nptr)->id << ":";

    for (n = 1; n <= m.size; n++) {
      ostrm.width(10);
      if ((m.lptr = ((m.nptr)->link)->search(n)) != NULL)   
	ostrm << (m.lptr)->value;
      else
	ostrm << ZERO;
    };
    ostrm << '\n';
    m.nptr = (m.nptr)->next;
  };

  if (ostrm.bad()) 
    error("error: incorrect output");    

  return ostrm;
} 

// ----< void Matrix :: deallocate_matrix() >---- //

// explicit matrix destructor
void Matrix :: deallocate_matrix()
{
  if (size != 0) {
    nptr = nhead;
    while (nptr != NULL) { 
      (nptr->link)->deallocate_llist();  
      nptr = nptr->next;
    };       
    nhead->deallocate_node();
  };
  size = 0;
  nhead = nptr = NULL;

  return;
}

// ----< void llist :: deallocate_llist() >---- //

// explicit link list destructor
void llist :: deallocate_llist()
{  
  Pointer<llist> ltmp1, ltmp2 = this;
  while (ltmp2 != NULL) {
    ltmp1 = ltmp2->next;	
    ltmp2.deleteptr(); 
    ltmp2 = ltmp1;
  };
  return;
}

// ----< void node :: deallocate_node() >---- //

// explicit graph node destructor
void node :: deallocate_node()
{
  Pointer<node> ntmp1, ntmp2 = this;
  while (ntmp2 != NULL) {
    ntmp1 = ntmp2->next;	
    ntmp2.deleteptr(); 
    ntmp2 = ntmp1;
  };
  return;
} 

// ----< Pointer<llist>  llist :: search(int id_index) >----//

// search for a connection in a link list
Pointer<llist>  llist :: search(int id_index)
{
  Pointer<llist> ltmp = this;
  while ((ltmp != NULL) && (ltmp->id != id_index)) ltmp = ltmp->next;
  return (ltmp);
}


/*             CSE 687 Final Project --- Pointer Class 

            Student Name: Gang Cheng SSN:223-61-3010               */

/*******************************************************************
 * final.cc ---- main program to test the member & frind functions *
 *               for class Matrix and class Pointer                *
 *******************************************************************/

#include <fstream.h>  // cin, cout, cerr
//#include "matrix.hh"  // class Matrix
//#include "pointer.hh" // class Pointer

//--- <void error(char *s)> ---//
/* error handler */
void error(char *s)
{
  cerr << s; 
  exit(1);
} 

//--- < main program > ---//
/* test the member and friend functions for class Matrix
   which are rewritten using class Pointer and class Pointer itself */
main(int argc, char *argv[])
{
  char fname1[10], fname2[10];

  switch (argc) {
  case 1:
    cerr << "Input file name for matrix 1: ";
    cin >> fname1;
    cerr << "Input file name for matrix 2: ";
    cin >> fname2;
    break;
  case 2:
    strcpy(fname1,argv[1]);
    cerr << "Input file name for matrix 2: ";
    cin >> fname2;
    break;
  case 3:
    strcpy(fname1,argv[1]);
    strcpy(fname2,argv[2]);
    break;
  default:
    error("error: incorrect command line input\n");
    break;
  };

  // open matrix input files
  ifstream fp1(fname1);
  ifstream fp2(fname2);
 
  // check i/o status
  if (!fp1) {
    cerr << "cannot open file: " << fname1 << "\n";
    exit(1);
  };
  if (!fp2) {
    cerr << "cannot open file: " << fname2 << "\n";
    exit(1);
  };

  // status information     
  cerr << "\nThese are files " << fname1 << " and " << fname2 
       << "\nProcessed by " <<  argv[0] << "\n\n";

  // Below are sample testing statments designed specifically
  //  for testing some functionlities of Pointer class which 
  //  are not found in the testing program for Matrix Class.
  //
  // The member function tested for Matrix Class are:
  //   1. void constructor Matrix();
  //   2. constructor Matrix(int);
  //   3. (default) destructor ~Matrix();
  //   4. operator = ;
  //   5. operator + ;
  //   6. operator * ;
  //   7. istream operator >> ;
  //   8. ostream operator << ;
  //   9. member function matrix_size();
  //   
  // The member functions tested for Pointer Class are:
  //   1. void constructor Pointer() ;
  //   2. promoted copy constructor Pointer(T*) ;
  //   3. (default) destructor ~Pointer() ;
  //   4. memory allocator newptr(int) ;
  //   5. memory deallocator deleteptr() ;
  //   6. operator * ;
  //   7. operator [] ;
  //   8. operator + ;
  //   9. operator ++ (prefix form) ;
  //  10. operator ++ (postfix form) ;
  //  11. operator -- (prefix form) ;
  //  12. operator=(Pointer&) ;
  //  
  // Both pointers to auto variable and to allocated memory
  //  are used in a mixed way to illustrate the nontrival
  //  functionality of Pointer class

  // define 3 pointers to matrix
  Pointer<Matrix> mptr,mp1, mp2;

  // mptr points to an array(2 elements) of matrix,
  // each matrix has size 0x0 at start
  mptr.newptr(2);

  // mp1, mp2 each points to a matrix of size 0x0 
  mp1.newptr(); mp2.newptr(); 

  // read 2 matrixes and make mptr pointing to them, 
  // the size and actual memory allocated for each matrix
  //  is determined from the input files 
  fp1 >> *(mptr.operator++(1)); // unfortunately, g++ on Cat does not
  fp2 >> *mptr;                 //  accept the postfix form invocation
                                //  of ++(or --), I have to use this 
                                //  awkward form instead
  // make mp1 and mp2 pointing to the same matrixes as mptr does
  mp1 = --mptr; 
  mp2 = mptr + 1;

  // define a new matrix (auto variable) with the same size as 2nd matrix 
  Matrix m(mptr->matrix_size());

  // m = matrix1 + matrix2
  m = mptr[0] + mptr[1];

  // (incorrectly) try to delete the memory mptr pointing to
  mptr.deleteptr();
  
  // output results
  // because mp1 and mp2 point to the same memory as mptr does,
  //  it illustrates the (incorrect) deletion of mptr's memory
  //  did not make trouble to other pointers to the same one
  cout << "\nm1 = \n"<<*mp1 << "\nm2 = \n" <<*mp2;
  cout << "\nm1 + m2 =  \n" << m;

  // following segements illustrate the mixing use of pointers
  //  pointing to allocated memory and to auto variables
  Pointer<Matrix> mp3(&m);
  *mp3 = (*mp1)*(*mp2);
  cout << "\nm1 * m2 =  \n" << m;

  Pointer<Matrix> mp4 = mp3;
  *mp4 = *mp1 + *mp4;  
  cout << "\nm1 + m1 * m2 =  \n" << *mp4;

  // more than once deletions of same memory is save now
  mp1.deleteptr(); // first deletion
  mp2.deleteptr();

  mp1.deleteptr(); // second deletion
  mp2.deleteptr();

  return 0;
}

// ****** END of the main program ******** //

/*********  Attached 2 sample matrix input files with the format:   ***********

   node_key_id1 : connected_node_id1 w weight connected_node_id2 w weight ... ;
   node_key_id1 : connected_node_id1 w weight connected_node_id2 w weight ... ;
     .......
     .......
   node_key_idn : connected_node_id1 w weight connected_node_id2 w weight ... ;

  *****************************************************************************/

// remove the following lines out into 2 seperate file2 (e.g. m1.dat, m2.dat)

/***** m1.dat *****

1 : 2w1.56 4w0.23; 
2 : ;
3 : 1w 2.4 3w 2.3;
4 : 2w23.4  1w4.45;  
5 : 2w1.56  4w0.23; 
6 : ;
7 : 3w2.4 7w2.3;
8 : 6w23.4  2w4.9;
9 : 1w2.4  8w0.1;
10 : 10w23.4;
11 : 2w-1.2 3w0.01;

****** end of m1.dat *******/

/***** m2.dat *****

1 : 1w2.45;
2 : 3w4.567 2w1.23;
3 : 6w23.4  8w4.452345; 
4 : 2w23.4  1w4.45;  
5 : 2w1.56 4w0.23; 
6 : 1w1.56 8w0.23; 
7 : ;
8 : 5w2.45 1w2.567 7w2.489;
9 : 1w1.2;
10 : ;
11 : 11w2.3;

****** end of m1.dat *******/

/***********************************************************
To run the test program, you can either enter 0 or 1 or 2 
input file name(s) on the command line, e.g. "final" or
"final m1.dat" or "final m1.dat m2.dat", in the case the program
needs more input file after starting the program, it will ask 
you to enter the file name(s). The output of the program will 
be to the standard output device(screen) by default. You can 
redirct the output to a file, e.g. "final >final.out" or
"final m1.dat >final.out" or "final m1.dat m2.dat >final.out".  
************************************************************/
 
